/* Ancillary functions for the edg-gridftp-base-* tools
 * CERN IT-DM/SMD (2008-02-25)
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <stdarg.h>

#include "edg-gridftp-base-common.h"

extern const char *prog;

int my_vfprintf(FILE *fp, const char *format, va_list ap) {
  va_list ap_copy;
  char *buf;
  char *c;
  FILE *dn;
  int len;

  dn = fopen("/dev/null", "w");
  if (dn == NULL) return -1;

  globus_libc_va_copy(ap_copy,ap);
  len = globus_libc_vfprintf(dn, format, ap_copy);
  va_end(ap_copy);
  fclose(dn);

  if (len<0) return -1;

  buf = malloc(len+1);
  if (buf == NULL) return -1;

  globus_libc_va_copy(ap_copy,ap);
  len = globus_libc_vsprintf(buf, format, ap_copy);
  va_end(ap_copy);

  if (len<0) {
    free(buf);
    return -1;
  }

  /* repace either \r\n or \n\r with just \n
   * e.g. \r\n\r -> \n\r
   *   or \r\n\r\n -> \n\n
   */
  for(c=buf; *c; ++c) {
    if ((c[0] == '\r' && c[1] == '\n') ||
        (c[0] == '\n' && c[1] == '\r')) {
      c[0] = '\n';
      memmove(&c[1], &c[2], strlen(c+1));
    }
  }

  /* remove any single \r at the beginning or end of the string */
  if (buf[0] == '\r') memmove(buf, &buf[1], strlen(buf));
  if (buf[strlen(buf)-1] == '\r') buf[strlen(buf)-1] = '\0';

  len = globus_libc_fprintf(fp, "%s", buf);

  free(buf);

  return (len);
}


int my_fprintf(FILE *fp, const char *format, ...) {
  va_list ap;
  int ret;

  va_start(ap, format);

  ret = my_vfprintf(fp, format, ap);

  va_end(ap);

  return ret;
}

void abort_handler(int signum) {
  fprintf(stderr, "%s: an error occurred and the program had to abort\n",prog);
  exit(FAIL);
}

void alarm_handler(int signum) {
  fprintf(stderr, "%s: timeout exceeded\n",prog);
  exit(FAIL);
}

void setup_timer_and_signals(int timeout) {
  struct sigaction act;

  memset(&act, '\0', sizeof(act));
  sigemptyset(&act.sa_mask);
  sigaddset(&act.sa_mask, SIGALRM);
  act.sa_handler = &abort_handler;
  sigaction(SIGABRT, &act, (struct sigaction *)NULL);

  memset(&act, '\0', sizeof(act));
  sigemptyset(&act.sa_mask);
  act.sa_handler = &alarm_handler;
  sigaction(SIGALRM, &act, (struct sigaction *)NULL);

  (void) alarm(timeout);
}

/*
 * A callback function for the GridFTP library. This is used to notify changes
 * for an operation request. The callback argument should be the monitor used
 * to synchronize the operation.  The handle is currently unused.  If there was
 * not a successful return, then the associated error text is printed to the
 * standard error stream.
 */ 
void callback(void *callback_arg, 
		     globus_ftp_client_handle_t *handle,
		     globus_object_t *error) {

  monitor_t *monitor = (monitor_t *) callback_arg;

  globus_mutex_lock(&monitor->mutex);

 /* there may also be an i/o callback in operation. If it exists check
  * if it has reported any error. If so we also indicate an error condition
  */
  if (monitor->monitor != NULL) {
    globus_mutex_lock(&monitor->monitor->mutex);
    if (monitor->monitor->error) {
      /* the data read callback has already noticed an error */
      monitor->done = GLOBUS_TRUE;
      monitor->error = GLOBUS_TRUE;
      globus_mutex_unlock(&monitor->monitor->mutex);
      globus_mutex_unlock(&monitor->mutex);
      return;
    }
    globus_mutex_unlock(&monitor->monitor->mutex);
  }

  if (error == GLOBUS_SUCCESS) {
    monitor->error = GLOBUS_FALSE;
  } else {
    monitor->error = GLOBUS_TRUE;
#if defined (GLOBUS_ERROR_TYPE_MULTIPLE)
    my_fprintf(stderr, "%s: error %s\n", prog, globus_error_print_friendly(error));
#else
    my_fprintf(stderr, "%s: error %s\n", prog, globus_object_printable_to_string(error));
#endif
  }

  monitor->done = GLOBUS_TRUE;
  globus_cond_signal(&monitor->cond);
  globus_mutex_unlock(&monitor->mutex);

} /* callback() */

/*
 * An optional callback function any necessary i/o operations that have to
 * happen after the operation request.
 * e.g. for read after a LIST operation request
 */ 
void data_callback(void *callback_arg, 
			  globus_ftp_client_handle_t *handle,
			  globus_object_t *error,
			  globus_byte_t *buffer,
			  globus_size_t length,
			  globus_off_t offset,
			  globus_bool_t eof) {

  monitor_t *monitor = (monitor_t *) callback_arg;

  globus_mutex_lock(&monitor->mutex);

  globus_mutex_lock(&monitor->monitor->mutex);

  if (monitor->monitor->error) {
    /* the operation callback has already noticed an error */
    monitor->done = GLOBUS_TRUE;
    monitor->error = GLOBUS_TRUE;
    globus_mutex_unlock(&monitor->monitor->mutex);
    globus_mutex_unlock(&monitor->mutex);
    return;
  }

  globus_mutex_unlock(&monitor->monitor->mutex);

  if (error == GLOBUS_SUCCESS) {
    monitor->error = GLOBUS_FALSE;
  } else {
    monitor->error = GLOBUS_TRUE;
    monitor->done = GLOBUS_TRUE;
#if defined (GLOBUS_ERROR_TYPE_MULTIPLE)
    my_fprintf(stderr, "%s: error %s\n", prog, globus_error_print_friendly(error));
#else
    my_fprintf(stderr, "%s: error %s\n", prog, globus_object_printable_to_string(error));
#endif
  }

  /* Print the result if there was no error. */
  if (!monitor->error) {
    buffer[length] = '\0';
    my_fprintf(stdout,"%s",buffer);
  }

  /* Check for end of file. */
  if (eof) monitor->done = GLOBUS_TRUE;

  globus_cond_signal(&monitor->cond);
  globus_mutex_unlock(&monitor->mutex);

  if (!eof) {
    globus_result_t gresult;
    
    /* Start another read operation. */
    gresult = globus_ftp_client_register_read(handle,
					      buffer,
					      BUFFER_SIZE-1,
					      &data_callback,
					      monitor);
    if (gresult != GLOBUS_SUCCESS) {
	my_fprintf(stderr, 
			    "%s: could not register read buffer\n",
			    prog);
    }
  }

} /* data_callback() */
